/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package StoneBreak.StoneBreakGame.Blocks;

import StoneBreak.StoneBreakGame.LevelManager;
import StoneBreak.StoneBreakGame.Globals;
import StoneBreak.StoneBreakGame.Globals.BLOCK_SPECIAL_ITEM;
import StoneBreak.StoneBreakGame.Paddle;
import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.GradientPaint;
import java.awt.Graphics2D;

/**
 * This is the base class for one block. It handles also the explosion of a block
 * as well as the special features that may be dropped.
 * @author SUN2K
 */
public abstract class BlockBaseType {

        /* position of block in pixels inside game area */
        protected int posX, posY;

        /* position of block in tile array */
        public int tileX, tileY;

        /* the general state of this block */
        protected enum BlockState { 
          EMPTY,        /* there is no block at this position at all */
          NORMAL,       /* block is available and waits to be hit */
          EXPLODING,    /* block was hit, but explosion animation is still running */
          SPECIALITEM };/* a special item is still visible, includes 'Exploding' */


        /* hold current block state */
        protected BlockState internalState;


        /* ------------------------------------------------------------------
         * Explosion related variables
         * ------------------------------------------------------------------ */

        /* array of actual particles */
        protected BlockParticle[] particles;

        /* current transparency value of particles */
        protected float particleAlpha;

        /* standard color for explosion, should be overwritten by sub classes */
        protected Color explosionColor = Color.LIGHT_GRAY;


        /* ------------------------------------------------------------------
         * Special item related variables
         * ------------------------------------------------------------------ */

        /* does this block contain a special item that will be activated if
         * the blocks explodes.
         */
        protected Globals.BLOCK_SPECIAL_ITEM specialItem;
        private double specialItemPosX, specialItemPosY;

        /* helper variable to handle recursive calls of the hit() function
         * in case of special item of exploding neighbour blocks
         */
        private static boolean NeighbourHitMutex = false;

        /**
         * Standard constructor.
         */
        public BlockBaseType()
        {
            internalState = BlockState.NORMAL;
            particles = new BlockParticle[Globals.BLOCK_NUM_PARTICLES];

            // temporary
            int tmp = (int)(Math.random() * 100);

            if (tmp <= 2)
                specialItem = Globals.BLOCK_SPECIAL_ITEM.PADDLE_SHOOTEX;
            else if (tmp <= 6)
                specialItem = Globals.BLOCK_SPECIAL_ITEM.PADDLE_SHOOT;
            else if (tmp <= 8)
                specialItem = Globals.BLOCK_SPECIAL_ITEM.EXTRA_LIVE;
            else if (tmp <= 11)
                specialItem = Globals.BLOCK_SPECIAL_ITEM.PADDLE_GROWTH;
            else if (tmp <= 13)
                specialItem = Globals.BLOCK_SPECIAL_ITEM.BALL_HIT_NEIGHBOURS;
            else
            {
                // no special item
                specialItem = Globals.BLOCK_SPECIAL_ITEM.NONE;
            }
            
            
        }

        /**
         * Here goes all the painting. Note that for state == NORMAL each
         * subclass has to implement its own drawing, for all other states
         * it should call this.
         * @param g Graphics object used for drawing
         */
        public void draw(Graphics2D g)
        {
            if (internalState == BlockState.NORMAL)
            {
                /* do nothing */
            }
            else if (internalState == BlockState.EXPLODING || internalState == BlockState.SPECIALITEM)
            {
                // process explosion
                if (particleAlpha > 0)
                {
                    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, particleAlpha));
                    for (int i = 0; i < Globals.BLOCK_NUM_PARTICLES; i++)
                    {
                        g.setColor(explosionColor);
                        g.fillOval((int)particles[i].x, (int)particles[i].y, BlockParticle.PARTICLESIZE, BlockParticle.PARTICLESIZE);
                        particles[i].x += particles[i].speedX;
                        particles[i].y += particles[i].speedY;
                        //particles[i].speedY += 0.1;
                    }
                    particleAlpha -= 0.012f;
                    g.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER));
                }

                // handle drawing of specialitem
                if (internalState == BlockState.SPECIALITEM)
                {
                    DrawSpecialItem(g);
                }

                /* state handling */
                if ((internalState == BlockState.EXPLODING) && (particleAlpha <= 0))
                    internalState = BlockState.EMPTY;
                if ((internalState == BlockState.SPECIALITEM) && (particleAlpha <= 0) &&
                    (specialItemPosY >= Globals.GAMEPANEL_HEIGHT))
                {
                    internalState = BlockState.EMPTY;
                }
            }
            else
            {
                /* should not happen */
            }
        }

        /**
         * Draw standard border around a block
         * @param g Graphics object used for drawing
         */
        public void drawBlockBorder(Graphics2D g)
        {
            g.setColor(Color.WHITE);
            g.drawLine(posX, posY, posX, posY + Globals.BLOCKSIZE_Y-1);     // left line
            g.drawLine(posX, posY, posX + Globals.BLOCKSIZE_X-1, posY);     // top line
            g.setColor(Color.BLACK);
            g.drawLine(posX, posY + Globals.BLOCKSIZE_Y-1, posX + Globals.BLOCKSIZE_X-1, posY + Globals.BLOCKSIZE_Y-1);
            g.drawLine(posX + Globals.BLOCKSIZE_X-1, posY, posX + Globals.BLOCKSIZE_X-1, posY + Globals.BLOCKSIZE_Y-1);
        }


        /**
         * Check if there is a block here actually
         * @return true if emtpty - in this case no need for drawing or collision handling required
         */
        public boolean IsEmpty()
        {
            return !(internalState == BlockState.NORMAL);
        }

        /**
         * Reset the block to its default state
         */
        public void resetNormalState()
        {
            internalState = BlockState.NORMAL;
        }

        /**
         * Check if the block should be drawn
         * @return true if draw method must be called
         */
        public boolean isVisible()
        {
            return (internalState != BlockState.EMPTY);
        }

        /**
         * Call this if ball hits the block. Should be overwritten by subclasses
         * for blocks that do not explode after one hit.
         */
        public void hit()
        {

            if (Paddle.getPaddleInstance().NeighbourHitFeatureActive && !NeighbourHitMutex)
            {
                NeighbourHitMutex = true;
                // the four direct neighbours should also be hit
                LevelManager lm = LevelManager.getLevelManager();
                // check left neightbour
                if (!(lm.getBlock(tileX-1, tileY) instanceof BlockTypeEmpty) && lm.getBlock(tileX-1, tileY).isVisible())
                    lm.getBlock(tileX-1, tileY).hit();
                // check right neightbour
                if (!(lm.getBlock(tileX+1, tileY) instanceof BlockTypeEmpty) && lm.getBlock(tileX+1, tileY).isVisible())
                    lm.getBlock(tileX+1, tileY).hit();
                // check top neightbour
                if (!(lm.getBlock(tileX, tileY-1) instanceof BlockTypeEmpty) && lm.getBlock(tileX, tileY-1).isVisible())
                    lm.getBlock(tileX, tileY-1).hit();
                // check bottom neightbour
                if (!(lm.getBlock(tileX, tileY+1) instanceof BlockTypeEmpty) && lm.getBlock(tileX, tileY+1).isVisible())
                    lm.getBlock(tileX, tileY+1).hit();

                NeighbourHitMutex = false;
            }

            finalHit();
        }

        /*
         * This is called to get the score of the block when it's hit.
         * Should be overwritten by subclasses.
         */
        public abstract int getScore();

        /**
         * Set x-postion of the block
         * @param newX x-coordinate in pixels
         */
        public void setPosX(int newX) { posX = newX; }
        /**
         * Set y-postion of the block
         * @param newY y-coordinate in pixels
         */
        public void setPosY(int newY) { posY = newY; }

        /**
         * Call this function if the block begins to explode.
         */
        public void finalHit()
        {

            // create the explosion particles
            for (int i = 0; i < Globals.BLOCK_NUM_PARTICLES; i++)
            {
                particles[i] = new BlockParticle(posX + (Globals.BLOCKSIZE_X >> 1), posY + (Globals.BLOCKSIZE_Y >> 1),
                        Math.random() - 0.5, (Math.random() - 0.5));
                particleAlpha = 1f;
            }
            if (specialItem == Globals.BLOCK_SPECIAL_ITEM.NONE)
            {
                // if this block contains no special item, just explosion
                internalState = BlockState.EXPLODING;
            }
            else
            {
                specialItemPosX = posX + (Globals.BLOCKSIZE_X >> 1) - (Globals.SPECIAL_ITEM_SIZEX>>1);
                specialItemPosY = posY + Globals.BLOCKSIZE_Y;
                internalState = BlockState.SPECIALITEM;
            }
        }


        
        /* ------------------------------------------------------------------
         * Special item
         * ------------------------------------------------------------------ */

        public double getSpecialItemPosX() { return specialItemPosX; }
        public double getSpecialItemPosY() { return specialItemPosY; }
        public BLOCK_SPECIAL_ITEM getSpecialItem() { return specialItem; }

        public boolean isSpecialItemActive()
        {
            return ( (internalState == BlockState.SPECIALITEM) &&
                     (specialItem != BLOCK_SPECIAL_ITEM.NONE));
        }

        private void DrawSpecialItem(Graphics2D g)
        {
            switch (specialItem)
            {
                case PADDLE_GROWTH:
                {
                    DrawSpecialItemBase(g, Color.GREEN, Color.GREEN.darker().darker());
                    g.fillRoundRect((int)(specialItemPosX + (Globals.SPECIAL_ITEM_SIZEX >> 1) - (Globals.SI_PADDLEGROWTH_WIDTH>>1)-1),
                               (int)(specialItemPosY + (Globals.SPECIAL_ITEM_SIZEY >> 1) - (Globals.SI_PADDLEGROWTH_HEIGHT>>1)-1),
                               Globals.SI_PADDLEGROWTH_WIDTH, Globals.SI_PADDLEGROWTH_HEIGHT, 6, 6);
                    DrawSpecialItemBorder(g);
                }
                break;

                case EXTRA_LIVE:
                {
                    DrawSpecialItemBase(g, Color.CYAN, Color.CYAN.darker().darker());
                    g.fillOval((int)(specialItemPosX + (Globals.SPECIAL_ITEM_SIZEX >> 1) - Globals.SI_EXTRALIFE_BALLRADIUS),
                               (int)(specialItemPosY + (Globals.SPECIAL_ITEM_SIZEY >> 1) - Globals.SI_EXTRALIFE_BALLRADIUS),
                               Globals.SI_EXTRALIFE_BALLRADIUS<<1, Globals.SI_EXTRALIFE_BALLRADIUS<<1);
                    DrawSpecialItemBorder(g);
                }
                break;

                case PADDLE_SHOOT:
                {
                    DrawSpecialItemBase(g, Globals.SI_PADDLESHOOT_COLOR, Color.RED.brighter());
                    DrawSpecialItemBorder(g);
                }
                break;

                case PADDLE_SHOOTEX:
                {
                    DrawSpecialItemBase(g, Globals.SI_PADDLESHOOTEX_COLOR, Color.RED.darker());
                    DrawSpecialItemBorder(g);
                }
                break;

                case BALL_HIT_NEIGHBOURS:
                {
                    DrawSpecialItemBase(g, Globals.SI_NEIGHBOURBLOCKSHIT_COLOR, Color.RED.darker());
                    DrawSpecialItemBorder(g);
                }
                break;


                default: break;     /* should not happen */
            }
        }

        /**
         * Move the special item of this block. Must be called every main cycle.
         */
        public void MoveSpecialItem()
        {
            specialItemPosY += Globals.SPECIAL_ITEM_SPEEDY;
        }

        /**
         * Remove special item. Called when paddle catches it or it moves out of the panel.
         */
        public void removeSpecialItem()
        {
            // set the specialitem out of the window, the actual state
            // handling will then happen in the draw method
            specialItemPosY = Globals.GAMEPANEL_HEIGHT + Globals.SPECIAL_ITEM_SIZEY;
        }

        private void DrawSpecialItemBorder(Graphics2D g)
        {
            g.setColor(Color.LIGHT_GRAY);
            g.drawRoundRect((int)specialItemPosX-1, (int)specialItemPosY-1,
                            Globals.SPECIAL_ITEM_SIZEX, Globals.SPECIAL_ITEM_SIZEY,
                            Globals.SPECIAL_ITEM_ROUNDRECT, Globals.SPECIAL_ITEM_ROUNDRECT);
        }

        private void DrawSpecialItemBase(Graphics2D g, Color color1, Color color2)
        {
            GradientPaint gp = new GradientPaint((int)specialItemPosX, (int)specialItemPosY, color1,
                                (int)specialItemPosX,
                                (int)specialItemPosY + Globals.SPECIAL_ITEM_SIZEY, color2, false);
            g.setPaint(gp);
            g.fillRoundRect((int)specialItemPosX, (int)specialItemPosY,
                    Globals.SPECIAL_ITEM_SIZEX, Globals.SPECIAL_ITEM_SIZEY,
                    Globals.SPECIAL_ITEM_ROUNDRECT, Globals.SPECIAL_ITEM_ROUNDRECT);
            g.setColor(Globals.BALL_COLOR);
        }

}
