/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package StoneBreak.StoneBreakGame;

/**
 *
 * @author SUN2K
 */


import StoneBreak.StoneBreakApplet;
import StoneBreak.StoneBreakGame.Blocks.*;
import java.awt.Color;
import java.awt.Graphics2D;
import java.io.DataInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

 public class LevelManager {

    /* Stonebreak level file constants */

    private static final int LEVEL_FILE_SIGNATURE	= 0x31464253;

    public static Color[] blockColors = { Color.BLUE, Color.RED, Color.GREEN, Color.YELLOW, Color.cyan,
                    Color.WHITE, Color.orange, Color.MAGENTA};

    /* total number of different levels */
    private int numOfLevels;

    /* number of current level; special case: 0 means no blocks at all */
    private int currentLevelNumber;

    /* hold structures of all available level */
    private GameLevel[] levels;

    /* total score achieved in game */
    private int totalScore;

    private static LevelManager instance = null;

    /**
     * Default constructor.
     */
    private LevelManager()
    {
        currentLevelNumber = 0;
        totalScore = 0;
    }

    /**
     * Get singleton instance of levelmanager.
     * @return levelmanager instance
     */
    public static LevelManager getLevelManager()
    {
        if (instance == null)
            instance = new LevelManager();
        return instance;
    }

    /**
     * Initialize all levels. Load and parse files and build level data structures.
     * @throws IOException if an error occurs while loading level files. If this happens,
     * this should be considered to be a fatal error and applet should exit.
     */
    public void InitLoadLevels() throws IOException
    {
        InputStream instream = null;
        boolean bAbortFileExistsLoop = false;
        int curFileNum = 0;

        // at first we have to examine how many files are able to be loaded
        while (!bAbortFileExistsLoop)
        {
            try
            {                
                instream = this.getClass().getResourceAsStream("/Level" + (curFileNum+1) + ".slf");
                if (instream == null)
                {
                    bAbortFileExistsLoop = true;
                    if (curFileNum == 0)
                    { // no level file existing at all -> abort
                        curFileNum = -1;
                    }
                }
                else
                {
                    // if we come to next line, file exists so go on with next file
                    curFileNum++;
                }
            }
            catch (Exception ex)
            {   // general error -> abort
                bAbortFileExistsLoop = true;
                curFileNum = -1;
            }
            finally
            {
                
                try
                {
                    if (instream != null)
                    {
                        instream.close();
                    }
                } catch (Exception ex)
                {
                    ex.printStackTrace();
                }
            }
        }

        if (curFileNum != -1)
        {   
            // no error occured during loading, init data structures
            numOfLevels = curFileNum;
            levels = new GameLevel[numOfLevels];
            for (int i = 0; i < numOfLevels; i++)
            {
                levels[i] = new GameLevel();
            }

            // load all levels
            for (int i = 0; i < numOfLevels; i++)
            {
                if (LoadLevelFromFile(this.getClass().getResourceAsStream("/Level" + (i+1) + ".slf"), levels[i]))
                {
                    // level loaded successfully from file - nothing to do yet
                }
                else
                {   // fatal error
                    throw new IOException();
                }
            }
        }
        else
        {   // fatal error
            throw new IOException();
        }
    }
    

    /**
     * Draw the current level (blocks and special items)
     * @param g The Graphics object to draw on.
     */
    public void DrawCurrentLevel(Graphics2D g)
    {
        Color oldColor = g.getColor();

        for (int y = Globals.NUM_BLOCKS_Y - 1; y >= 0; y--)
        {
            for (int x = Globals.NUM_BLOCKS_X - 1; x >= 0; x--)
            {
                if (levels[currentLevelNumber].levelData[y][x].isVisible())
                {
                    levels[currentLevelNumber].levelData[y][x].draw(g);
                }
            }
        }

        g.setColor(oldColor);
    }

    /*
     * Recover the initial state of the cuurent level.
     */
    public void ReInitCurrentLevel() { ReInitLevel(currentLevelNumber); }

    /**
     * Recover the intital state of the level
     * @param levelNum number of level to recover. First level is 0, second is 1 etc.
     */
    public void ReInitLevel(int levelNum)
    {
        GameLevel level = levels[levelNum];

        for (int row = 0; row < Globals.NUM_BLOCKS_Y; row++)
        {
            for (int column = 0; column < Globals.NUM_BLOCKS_X; column++)
            {
                // an empty block is statically empty and must not be changed
                if ( !(level.levelData[row][column] instanceof BlockTypeEmpty))
                {
                    level.levelData[row][column].resetNormalState();
                }
            }
        }
    }

    /**
     * Check if the current level is finished.
     * @return true if current level is finished, otherwise false.
     */
    public boolean IsCurrentLevelFinsihed() { return IsLevelFinished(currentLevelNumber); }

    /**
     * Check if the given level is finished.
     * @param levelNum number of level to check if it's solved.
     * @return true if level is finished, otherwise false.
     */
    public boolean IsLevelFinished(int levelNum)
    {
        boolean result = true;

        GameLevel level = levels[levelNum];
        for (int row = 0; row < Globals.NUM_BLOCKS_Y; row++)
        {
            for (int column = 0; column < Globals.NUM_BLOCKS_X; column++)
            {
                if ( !((level.levelData[row][column].IsEmpty()) ||
                      (level.levelData[row][column] instanceof BlockTypeUnbreakable)) )
                {
                    // this block was not hit and it can be destroyed, so level is not finished
                    result = false;
                    // abort loop
                    column = Globals.NUM_BLOCKS_X;
                    row = Globals.NUM_BLOCKS_Y;
                }
            }
        }

        return result;
    }

    /**
     * Set the levelManager to the next level = increases currentlevel.
     * @return true if next level is accepted; false if the current level was
     * already the last level, in this case nothing is done.
     */
    public boolean SwitchToNextLevel()
    {
        totalScore += levels[currentLevelNumber].getCurrentLevelScore();
        return SwitchToLevel(currentLevelNumber + 1);
    }

    /**
     * Set the levelManager to the given level
     * @param newLevel number of level to be set to active
     * @return true if level is switched, otherwise false
     */
    public boolean SwitchToLevel(int newLevel)
    {
        levels[currentLevelNumber].resetLevelScore();
        if (newLevel >= numOfLevels || newLevel < 0)
        {
            // we are currently at the last level or level is invalid
            return false;
        }
        else
        {
            currentLevelNumber = newLevel;
            return true;
        }
    }

    /**
     * Returns the total number of levels
     * @return number of levels
     */
    public int getNumOfLevels() { return numOfLevels; }

    /**
     * Returns the number of the current level
     * @return current level number, first level is 0.
     */
    public int getCurLevelNum() { return currentLevelNumber; }

    /**
     * Returns the level data of the current level
     * @return GameLevel structure holding the actual level
     */
    public GameLevel getCurLevel() { return levels[currentLevelNumber]; }

    /**
     *
     * @return the total game score
     */
    public int getTotalScore()
    {
        return totalScore + levels[currentLevelNumber].getCurrentLevelScore();
    }

    /**
     * set the total game score back to zero.
     */
    public void resetTotalScore()
    {
        totalScore = 0;
    }

    /**
     * Increase the total game score
     * @param score amount to increase the score.
     */
    public void addScore(int score)
    {
        totalScore += score;
    }

    public BlockBaseType getBlock(int x, int y)
    {
        if ( (y >= 0) && (y < Globals.NUM_BLOCKS_Y) && (x >= 0) && (x < Globals.NUM_BLOCKS_X))
            return levels[currentLevelNumber].levelData[y][x];
        else
            return new BlockTypeEmpty();
    }

    /**
     * Load level from file into level datastructure.
     * @param levelUrl URL to level file to load
     * @param level where level data is stored into
     * @return true if loading was successful, false if an error occured.
     */
    private boolean LoadLevelFromFile(InputStream instream, GameLevel level)
    {
        DataInputStream dataIn = null;
        boolean result = true;
        int expFileLength = 4 + 1 + 1 + 2 + (Globals.NUM_BLOCKS_X * Globals.NUM_BLOCKS_Y);
        byte[] databytes = new byte[expFileLength];
        
        try
        {
            dataIn = new DataInputStream(instream);
            if (dataIn.read(databytes) != expFileLength)
            {   // File invalid
                return false;
            }
            else
            {
                // consistency checks
                if ( ((int)(databytes[0] | (databytes[1] << 8) | (databytes[2] << 16) | (databytes[3] << 24) ) != LEVEL_FILE_SIGNATURE) ||
                     (databytes[4] != Globals.NUM_BLOCKS_X) ||
                     (databytes[5] != Globals.NUM_BLOCKS_Y) )
                {   // File invalid
                    return false;
                }
                else
                {   // read actual level data
                    byte curBlock;
                    for (int row = 0; row < Globals.NUM_BLOCKS_Y; row++)
                    {
                        for (int column = 0; column < Globals.NUM_BLOCKS_X; column++)
                        {
                            curBlock = databytes[6 + row * Globals.NUM_BLOCKS_X + column];

                            switch (curBlock)
                            {
                                case 0x01:
                                    level.levelData[row][column] = new BlockTypeNormalGradient();
                                    break;

                                case 0x02:
                                    level.levelData[row][column] = new BlockTypeStandard(BlockTypeStandard.BLOCKTYPE_STANDARD.BLUE);
                                    break;

                                case 0x03:
                                    level.levelData[row][column] = new BlockTypeStandard(BlockTypeStandard.BLOCKTYPE_STANDARD.GREEN);
                                    break;

                                case 0x10:
                                    level.levelData[row][column] = new BlockType2Hits();
                                    break;

                                case 0x11:
                                    level.levelData[row][column] = new BlockType3Hits();
                                    break;

                                case 0x12:
                                    level.levelData[row][column] = new BlockTypeUnbreakable();
                                    break;

                                case 0x00:
                                default:
                                    level.levelData[row][column] = new BlockTypeEmpty();
                                    break;
                            }

                            level.levelData[row][column].setPosX(column * Globals.BLOCKSIZE_X);
                            level.levelData[row][column].setPosY(Globals.BLOCK_AREA_TOP + row * Globals.BLOCKSIZE_Y);
                            level.levelData[row][column].tileX = column;
                            level.levelData[row][column].tileY = row;
                        }
                    }

                    // finally check checksum
                    int chksum = CalcLevelDataChecksum(databytes, 6, Globals.NUM_BLOCKS_X * Globals.NUM_BLOCKS_Y);
                    if ((short)(chksum & 0x0000FFFF) !=
                       (short)((  ((short)(databytes[expFileLength-2]) & 0x00FF) |
                                  (((short)(databytes[expFileLength-1]) & 0x00FF) << 8) ) & 0x0000FFFF) )
                    {
                        return false;
                    }
                }
            }
            
        }
        catch (Exception ex)
        {
            result = false;
            ex.printStackTrace();
        }

        finally
        {
            if (dataIn != null)
            {
                try
                {
                    dataIn.close();
                } catch (Exception ex)
                {
                    ex.printStackTrace();
                }
            }
        }

        return result;
    }

    /**
     * Calculates the chksum of the level data. Note that an unsigned 16bit is used
     * for the calculation, but since Java's short is signed we use an int (32bit)
     * and take only the lower 16bits into account.
     * @param databytes array hold the level
     * @param startIndex start index in the array to begin with calculation
     * @param length how many bytes from startIndex should be processed
     * @return
     */
    private int CalcLevelDataChecksum(byte[] databytes, int startIndex, int length)
    {
        int chksum = 0;
	for (int i = startIndex; i < startIndex+length; i++)
	{
		chksum += (databytes[i] & 0xFFFF);
	}
	return ((int)(~chksum) & 0xFFFF);
    }





    /* --- XML Parsing Stuff --- */
    /*
    private boolean LoadLevelFromFileXML(int levelNum, Level level)
    {
        org.w3c.dom.Document doc;
        NodeList nlTmp, nlTmp2;
        int intTmp;

        // variables for temporary level data
        int levelSizeX, levelSizeY;
        boolean result = true;
        
        try
        {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            doc = db.parse(getClass().getResourceAsStream("Levels/Level1.xml"));

            System.out.println("Root element " + doc.getDocumentElement().getNodeName());

            // get size of level

            nlTmp = doc.getElementsByTagName("NumBlocksInX");
            nlTmp2 = ((Element)nlTmp.item(0)).getChildNodes();
            levelSizeX = Integer.parseInt(((Node)nlTmp2.item(0)).getNodeValue());

            nlTmp = doc.getElementsByTagName("NumBlocksInY");
            nlTmp2 = ((Element)nlTmp.item(0)).getChildNodes();
            levelSizeY = Integer.parseInt(((Node)nlTmp2.item(0)).getNodeValue());

            System.out.println("X: " + levelSizeX + "  Y: " + levelSizeY);

            // get the rows
            NodeList nlRows = doc.getElementsByTagName("Row");
            if (nlRows.getLength() != Globals.NUM_BLOCKS_Y)
            {
                System.out.println("Number of rows in level file does not match");
                result = false;
            }


            for (int row = 0; row < nlRows.getLength(); row++)
            {
                System.out.println("Row: " + row);
                
                NodeList nlBlocks = ((Element)nlRows.item(row)).getElementsByTagName("Block");
                if (nlBlocks.getLength() != Globals.NUM_BLOCKS_X)
                {
                    System.out.println("Number of blocks in row " + row + " in level file does not match");
                    result = false;
                }

                for (int column = 0; column < nlBlocks.getLength(); column++)
                {
                    System.out.println("Block: " + column);

                    NodeList nlColor = ((Element)nlBlocks.item(column)).getElementsByTagName("ColorIndex");
                    intTmp = Integer.parseInt(nlColor.item(0).getChildNodes().item(0).getNodeValue());

                    int r = intTmp % 7;
                    
                    if (r == 0)
                    {
                        level.levelData[row][column] = new BlockTypeNormalGradient();
                    }
                    else if (r == 1)
                    {
                        level.levelData[row][column] = new BlockTypeStandard(1);
                    }
                    else if (r == 2)
                    {
                        level.levelData[row][column] = new BlockTypeStandard(2);
                    }
                    else if (r == 3)
                    {
                        level.levelData[row][column] = new BlockType2Hits();
                    }
                    else if (r == 4)
                    {
                        level.levelData[row][column] = new BlockTypeUnbreakable();
                    }
                    else if (r == 5)
                    {
                        level.levelData[row][column] = new BlockType3Hits();
                    }
                    else
                        // TODO: need special class for empty block
                        level.levelData[row][column] = new BlockTypeEmpty();

                    //level.levelData[row][column].blockColorIndex = intTmp;
                    if (intTmp == -1)
                    {
                        level.levelData[row][column].setInitialEmptyState();
                    }
                    //level.levelData[row][column].isEmpty = (intTmp == -1);
                    //level.levelData[row][column].isActive = true;

                    level.levelData[row][column].setPosX(column * Globals.BLOCKSIZE_X);
                    level.levelData[row][column].setPosY(Globals.BLOCK_AREA_TOP + row * Globals.BLOCKSIZE_Y);
                    level.levelData[row][column].tileX = column;
                    level.levelData[row][column].tileY = row;

                    
                }

            }
           

        }
        catch (ParserConfigurationException ex)
        {
            Logger.getLogger(LevelManager.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
            result = false;
        }
        catch (SAXException ex) {
            Logger.getLogger(LevelManager.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
            result = false;
        }
        catch (IOException ex) {
            Logger.getLogger(LevelManager.class.getName()).log(java.util.logging.Level.SEVERE, null, ex);
            result = false;
        }

        return result;
    }
    */
    
}
